@@ -1,4 +1,4 @@
-/* Copyright (c) 2007-2010 Sam Trenholme and others
+/* Copyright (c) 2007-2012 Sam Trenholme and others
  *
  * TERMS
  *
@@ -56,7 +56,7 @@ extern int resurrections;
 extern int min_bind;
 extern int num_ports;
 extern int num_retries;
-extern int deliver_all;
+extern int32_t max_ttl;
 
 /* Other mararc parameters */
 extern dwd_dict *blacklist_dict;
@@ -180,8 +180,9 @@ void make_remote_connection(int32_t n, unsigned char *packet, int len,
      dw_str *query, SOCKET x_socket_num) {
         sockaddr_all_T server;
         SOCKET s = 0;
-        int32_t rnum = -1;
+        int_fast32_t rnum = -1;
         ip_addr_T addr = {0,{0,0},0,0};
+        int counter = 0;
         socklen_t inet_len = sizeof(struct sockaddr_in);
 
         if(rem[n].socket != x_socket_num || /* Already used (sanity check) */
@@ -196,7 +197,19 @@ void make_remote_connection(int32_t n, unsigned char *packet, int len,
         }
 
         /* Get IP of remote server and open a socket to them */
-        addr = get_upstream_ip(query,n);
+        for(counter = 0; counter < 7; counter++) { /* Don't stop on bad IP */
+                addr = get_upstream_ip(query,n);
+                /* Make sure RD is appropriately set or clear */
+                if(rem[n].is_upstream == 1) {
+                        packet[2] |= 0x01; /* Set RD */
+                } else {
+                        packet[2] &= 0xfe; /* Clear RD */
+                }
+                if(addr.len > 0) {
+                        break;
+                }
+        }
+        dw_log_ip("Making connection to IP",&addr,128);
         if(addr.len == 0) { /* Failed to get upstream IP */
                 return;
         } else if(addr.len == 4 || addr.len == 16) { /* IPv4/IPv6 IP */
@@ -272,6 +285,7 @@ int find_inflight_query(unsigned char *a, int len) {
         }
 
         query = dw_get_dname_type(a,12,len);
+        dwc_lower_case(query);
         if(query == 0) {
                 goto catch_find_inflight_query;
         }
@@ -319,6 +333,7 @@ void make_new_udp_connect(int b, unsigned char *a, int len, int num_alloc) {
                 return;
         }
         rem[b].query = dw_get_dname_type(a,12,len);
+        dwc_lower_case(rem[b].query);
         rem[b].local = malloc(num_alloc * sizeof(local_T *));
         if(rem[b].local == 0) {
                 return;
@@ -347,9 +362,10 @@ void make_new_udp_connect(int b, unsigned char *a, int len, int num_alloc) {
  * to be passed to the function. */
 int forward_local_udp_packet(SOCKET sock, int32_t local_id,
      ip_addr_T *from_ip, uint16_t from_port, unsigned char *a,
-     int len, int tcp_num) {
+     int len, int tcp_num, dw_str *orig_query) {
         int32_t b = 0;
         int num_alloc = 0;
+        ip_addr_T null_addr = {0,{0,0},0,0};
 
         num_alloc = key_n[DWM_N_max_inflights];
         if(num_alloc < 1) {
@@ -380,6 +396,10 @@ int forward_local_udp_packet(SOCKET sock, int32_t local_id,
         }
 
         rem[b].local[rem[b].num_locals - 1] = malloc(sizeof(local_T));
+        if(rem[b].local[rem[b].num_locals - 1] != 0) {
+                rem[b].local[rem[b].num_locals - 1]->orig_query = 0;
+                rem[b].local[rem[b].num_locals - 1]->action = 0;
+        }
         if(rem[b].socket == INVALID_SOCKET ||
            rem[b].local[rem[b].num_locals - 1] == 0) {
                 reset_rem(b);
@@ -389,11 +409,14 @@ int forward_local_udp_packet(SOCKET sock, int32_t local_id,
         rem[b].local[rem[b].num_locals - 1]->tcp_num = tcp_num;
         if(from_ip != 0) {
                 rem[b].local[rem[b].num_locals - 1]->ip = *from_ip;
+        } else {
+                rem[b].local[rem[b].num_locals - 1]->ip = null_addr;
         }
         rem[b].local[rem[b].num_locals - 1]->port = from_port;
         rem[b].local[rem[b].num_locals - 1]->local_id = local_id;
         rem[b].local[rem[b].num_locals - 1]->glueless_type = 0;
         rem[b].local[rem[b].num_locals - 1]->glueless_conn = -1;
+        rem[b].local[rem[b].num_locals - 1]->orig_query = dw_copy(orig_query);
         rem[b].die = get_time() + ((int64_t)timeout_seconds << 8);
         return 0;
 }
@@ -451,9 +474,12 @@ dw_str *make_dns_packet(dw_str *question, dw_str *answer, int32_t id) {
         dw_str *out = 0;
 
         is_nxdomain = dw_pop_u8(answer);
-        arcount = dw_pop_u16(answer);
-        nscount = dw_pop_u16(answer);
-        ancount = dw_pop_u16(answer);
+        if(is_nxdomain != TYPE_TRUNCATED &&
+                        is_nxdomain != TYPE_TRUNCATED_NXDOMAIN) {
+                arcount = dw_pop_u16(answer);
+                nscount = dw_pop_u16(answer);
+                ancount = dw_pop_u16(answer);
+        }
 
         if(is_nxdomain == 0 || is_nxdomain == 2) {
                 /* 0x8180: QR = 1; Opcode = 0; AA = 0; TC = 0; RD = 1; RA = 1;
@@ -462,13 +488,23 @@ dw_str *make_dns_packet(dw_str *question, dw_str *answer, int32_t id) {
         } else if(is_nxdomain == 1) {
                 /* Same header as before, but with RCODE of "name error" */
                 out = make_dns_header(id,0x8183,ancount,nscount,arcount);
+        } else if(is_nxdomain == TYPE_TRUNCATED) {
+                /* Set TC to 1 */
+                out = make_dns_header(id,0x8380,0,0,0);
+        } else if(is_nxdomain == TYPE_TRUNCATED_NXDOMAIN) {
+                /* TC 1; RCODE "name error" */
+                out = make_dns_header(id,0x8383,0,0,0);
         } else {
                 goto catch_make_dns_packet;
         }
 
         if(dw_append(question,out) == -1 ||
-           dw_put_u16(out,1,-1) == -1 || /* QCLASS: 1 */
-           dw_append(answer,out) == -1) {
+           dw_put_u16(out,1,-1) == -1 /* QCLASS: 1 */ ) {
+                goto catch_make_dns_packet;
+        }
+        if(is_nxdomain != TYPE_TRUNCATED &&
+                        is_nxdomain != TYPE_TRUNCATED_NXDOMAIN &&
+                        dw_append(answer,out) == -1) {
                 goto catch_make_dns_packet;
         }
 
@@ -486,12 +522,13 @@ catch_make_dns_packet:
  * from the cache */
 int get_reply_from_cache(dw_str *query, sockaddr_all_T *client,
                          SOCKET sock, int32_t id, int resurrect,
-                         int tcp_num) {
+                         int tcp_num, dw_str *orig_query,
+                         ip_addr_T *from_ip) {
         dw_str *value = 0; /* Element in cache */
         dw_str *comp = 0; /* Compressed DNS packet */
         dw_str *packet = 0;
         socklen_t c_len = sizeof(struct sockaddr_in);
-        int ret = -1, type = 0;
+        int ret = -1, type = 0, cache_type = 0;
 
 #ifdef IPV6
         if (client->Family == AF_INET6) {
@@ -499,13 +536,33 @@ int get_reply_from_cache(dw_str *query, sockaddr_all_T *client,
         }
 #endif
 
-        dw_log_dwstr("Sending reply for query ",query,100);
+        dw_log_dwstr("Looking in cache for query ",query,100);
         type = dw_fetch_u16(query,-1);
-        if(type != 255) { /* RR_ANY; dwc_process munges ANY replies */
-                dwc_process(cache,query,3); /* RR rotation, TTL aging, etc. */
+        if(type == RR_MX && key_n[DWM_N_reject_mx] != 0) {
+                dw_log_ip(
+        "Attempt to get MX record (possible spam zombie) from IP",from_ip,3);
+                return 1;
         }
+        dwc_process(cache,query,3); /* RR rotation, TTL aging, etc. */
         value = dwh_get(cache,query,resurrect,1);
-        comp = dwc_compress(query,value);
+        if(value == 0) {
+                dw_log_dwstr("Nothing found for ",query,100);
+                goto catch_get_reply_from_cache;
+        }
+        cache_type = dw_fetch_u8(value,-1);
+        if(cache_type != TYPE_TRUNCATED &&
+                        cache_type != TYPE_TRUNCATED_NXDOMAIN) {
+                comp = dwc_compress(query,value);
+        } else {
+                /* Immediately zap truncated from cache when fetched */
+                dwh_zap(cache,query,0,1);
+                if(client == 0) { /* DNS-over-TCP */
+                        ret = 2;
+                        goto catch_get_reply_from_cache;
+                }
+                comp = dw_copy(value);
+        }
+
         if(comp == 0) {
                 goto catch_get_reply_from_cache;
         }
@@ -517,7 +574,7 @@ int get_reply_from_cache(dw_str *query, sockaddr_all_T *client,
         }
 
         dw_log_dwstr_str("Fetching ",query," from cache",100);
-        packet = make_dns_packet(query,comp,id);
+        packet = make_dns_packet(orig_query,comp,id);
         if(packet == 0) {
                 goto catch_get_reply_from_cache;
         }
@@ -549,20 +606,21 @@ catch_get_reply_from_cache:
    expired element with the data we want. */
 void try_forward_local_udp_packet(SOCKET sock, int32_t local_id,
      ip_addr_T *from_ip, uint16_t from_port, unsigned char *packet,
-     int len, sockaddr_all_T *client,dw_str *query, int tcp_num) {
+     int len, sockaddr_all_T *client,dw_str *query, int tcp_num,
+     dw_str *orig_query) {
 
         /* If not cached, get a reply that we will cache and send back to
          * the client */
         if(forward_local_udp_packet(sock,local_id,from_ip,from_port,
-                                    packet,len,tcp_num) != -1) {
+                                    packet,len,tcp_num,orig_query) != -1) {
                 return; /* Success! */
         }
 
         /* OK, at this point it failed so we'll see if we get a
          * "resurrected" cache entry */
         if(resurrections == 1 &&
-           get_reply_from_cache(query,client,sock,local_id,1,tcp_num)
-             == 1) {
+           get_reply_from_cache(query, client, sock, local_id, 1, tcp_num,
+                        orig_query, from_ip) == 1) {
                 dw_log_string("Resurrected from cache",11);
                 return; /* Resurrected entry; we're done */
         }
@@ -581,7 +639,8 @@ void get_local_udp_packet(SOCKET sock) {
         ip_addr_T from_ip = {0,{0,0},0,0};
         uint16_t from_port = 0;
         int32_t local_id = -1;
-        dw_str *query = 0;
+        dw_str *query = 0, *orig_query = 0;
+        int_fast32_t qtype = 0;
 #ifdef VALGRIND_NOERRORS
         memset(packet,0,522);
 #endif /* VALGRIND_NOERRORS */
@@ -591,6 +650,10 @@ void get_local_udp_packet(SOCKET sock) {
         len = recvfrom(sock,(void *)packet,520,0,(struct sockaddr *)&client,
                        &c_len);
 
+        if(len < 12) {
+                goto catch_get_local_udp_packet;
+        }
+
         from_port = get_from_ip_port(&from_ip,&client);
 
         if(check_ip_acl(&from_ip) != 1) { /* Drop unauthorized packets */
@@ -598,6 +661,24 @@ void get_local_udp_packet(SOCKET sock) {
         }
 
         local_id = get_dns_qid(packet, len, 2);
+        if(local_id == -2) { /* Immediate NOTIMPL for EDNS requests w/ OPT */
+#ifdef STRICT_RFC2671_COMPLIANCE
+                packet[2] = 0x81; /* QR = 1; Op = 0; AA = 0; TC = 0; RD = 1 */
+                packet[3] = 0x84; /* RA = 1; Z = 0; RCODE = "notimpl" (4) */
+                sendto(sock,(void *)packet,len,0,(struct sockaddr *)&client,
+                                c_len);
+                goto catch_get_local_udp_packet;
+#else /* STRICT_RFC2671_COMPLIANCE */
+                query = dw_get_dname_type(packet,12,len);
+                if(query == 0) {
+                        goto catch_get_local_udp_packet;
+                }
+                len = query->len + 14;
+                packet[11] = 0; /* We no longer have final question */
+                dw_destroy(query);
+                local_id = get_dns_qid(packet, len, 2);
+#endif /* STRICT_RFC2671_COMPLIANCE */
+        }
         if(local_id < 0 || len < 13) { /* Invalid remote packet */
                 goto catch_get_local_udp_packet;
         }
@@ -605,24 +686,59 @@ void get_local_udp_packet(SOCKET sock) {
         /* See if we have something in the cache; destroyed at end of
          * function */
         query = dw_get_dname_type(packet,12,len);
+        qtype = dw_fetch_u16(query,-1);
+
+        /* Reject PTR or AAAA queries if not wanted */
+        if((qtype == 28 /* AAAA */ && key_n[DWM_N_reject_aaaa] == 1) ||
+           (qtype == 12 /* PTR */ && key_n[DWM_N_reject_ptr] == 1)) {
+                unsigned char *answer;
+                answer = make_synth_not_there_answer(packet,&len,0);
+
+                /* Flag this as an answer */
+                answer[2] |= 0x80;
+
+                /* Flag RA bit because, well, recursion is available */
+                answer[3] |= 0x80;
+
+                /* One "NS" record; no other records */
+                answer[9] = 1;
+                answer[6] = answer[7] = answer[8] = answer[10] = answer[11] =0;
+                /* Copy over ID */
+                sendto(sock,(void *)answer,len+40,0,(struct sockaddr *)&client,
+                                c_len);
+                free(answer);
+                goto catch_get_local_udp_packet;
+        }
+
+        orig_query = dw_copy(query);
+        dwc_lower_case(query);
         dw_log_dwstr("Got DNS query for ",query,100); /* Log it */
         if(query == 0) {
                 goto catch_get_local_udp_packet;
         }
-        if(get_reply_from_cache(query,&client,sock,local_id,0,-1)
-                == 1) {
+        if(get_reply_from_cache(query, &client, sock, local_id, 0, -1,
+                        orig_query, &from_ip) == 1) { /* Is answer in cache */
                 goto catch_get_local_udp_packet; /* In cache; we're done */
         }
+        if(dwx_cname_in_cache(orig_query, query, &client, &from_ip, local_id,
+                        sock, from_port) == 1) { /* CNAME refer in cache? */
+                goto catch_get_local_udp_packet;
+        }
 
         /* Nothing in cache; lets try sending the packet upstream */
         try_forward_local_udp_packet(sock,local_id,&from_ip,from_port,
-                        packet,len,&client,query,INVALID_SOCKET);
+                        packet,len,&client,query,INVALID_SOCKET,
+                        orig_query);
 
 catch_get_local_udp_packet:
         if(query != 0) {
                 dw_destroy(query);
                 query = 0;
         }
+        if(orig_query != 0) {
+                dw_destroy(orig_query);
+                query = 0;
+        }
 }
 
 /* Forward a remote reply back to the client over UDP */
@@ -669,7 +785,7 @@ void forward_remote_reply(unsigned char *packet, size_t len, remote_T *r_ip,
 
 /* Add a reply we have received from the remote (upstream) DNS server to
  * the cache */
-int cache_dns_reply(unsigned char *packet, int count, int b) {
+int cache_dns_reply(unsigned char *packet, int count, int b, int truncated) {
         int32_t ttl = 60;
         int32_t ancount = 0;
         int is_nxdomain = 0;
@@ -678,40 +794,62 @@ int cache_dns_reply(unsigned char *packet, int count, int b) {
         int ret = -1, qtype = -1;
 
         question = dw_get_dname_type(packet,12,count);
+        dwc_lower_case(question);
         qtype = dw_fetch_u16(question,-1);
         dw_log_dwstr("Caching a reply for query ",question,1000);
-        if((packet[3] & 0x0f) == 3) {
+        if((packet[3] & 0x0f) == 2) { /* Server FAIL */
+                ret = -1; /* Bad return value; do not cache */
+                goto catch_cache_dns_reply;
+        } else if((packet[3] & 0x0f) == 3) { /* Name error/NXDOMAIN */
                 is_nxdomain = 1;
         }
-        answer = dw_packet_to_cache(packet,count,is_nxdomain);
-        decomp = dwc_decompress(question,answer);
-        if(dwc_has_bad_ip(decomp,blacklist_dict) ||
-           (qtype == 28 /* AAAA */ && key_n[DWM_N_reject_aaaa] == 1)) {
-                ret = -2; /* Tell caller we need synth "not there" */
-                goto catch_cache_dns_reply;
-        }
-        ancount = dw_cachepacket_to_ancount(answer);
-        if(ancount == 0) {
-                ancount = 32; /* So we can correctly cache negative answers */
-        }
+        if(truncated == 1) {
+                is_nxdomain += 3;
+                answer = dw_create(2);
+                if(dw_put_u8(answer, is_nxdomain, 0) == -1) {
+                        goto catch_cache_dns_reply;
+                }
+                dwh_add(cache,question,answer,7,1);
+                ret = 1;
+        } else {
+                answer = dw_packet_to_cache(packet,count,is_nxdomain);
+                decomp = dwc_decompress(question,answer);
+                if(decomp == 0) {
+                        goto catch_cache_dns_reply;
+                }
+                if(dwc_has_bad_ip(decomp,blacklist_dict)) {
+                        ret = -2; /* Tell caller we need synth "not there" */
+                        goto catch_cache_dns_reply;
+                }
+                ancount = dw_cachepacket_to_ancount(answer);
+                if(ancount == 0) {
+                        ancount = 32; /* Correct negative answer caching */
+                }
 
-        if(question == 0 || answer == 0 || ancount == -1) {
-                goto catch_cache_dns_reply;
-        }
+                if(question == 0 || answer == 0 || ancount == -1) {
+                        goto catch_cache_dns_reply;
+                }
 
-        ttl = dw_get_a_dnsttl(answer,0,31536000,ancount);
+                ttl = dw_get_a_dnsttl(answer,0,31536000,ancount);
+                if(ttl == -1 && decomp != 0 && decomp->len == 7) {
+                        ttl = 30; /* Special case: Blank reply upstream */
+                }
+                if(ttl == -1) {
+                        goto catch_cache_dns_reply;
+                }
+                if(ttl < 30) {
+                        ttl = 30;
+                }
+                if(ttl > max_ttl) {
+                        ttl = max_ttl;
+                }
 
-        if(ttl == -1) {
-                goto catch_cache_dns_reply;
+                /* Routines in DwRecurse.c process the packet and let us know
+                 * what kind of packet we got upstream (so we know how to
+                 * continue)
+                 */
+                ret = dwx_cache_reply(cache,question,decomp,ttl,b);
         }
-        if(ttl < 30) {
-                ttl = 30;
-        }
-
-        /* Have routines in DwRecurse.c process the packet and let us know
-         * what kind of packet we got upstream (so we know how to continue)
-         */
-        ret = dwx_cache_reply(cache,question,decomp,ttl,b);
 
 catch_cache_dns_reply:
         if(question != 0) {
@@ -743,7 +881,7 @@ int verify_dns_packet(int b, unsigned char *packet, int len) {
                 goto catch_verify_dns_packet;
         }
 
-        /* Should we make this case-insensitive? Probably not. */
+        dwc_lower_case(question); /* Case-insensitive comparison */
         if(dw_issame(question,rem[b].query) != 1) {
                 goto catch_verify_dns_packet;
         }
@@ -759,7 +897,8 @@ catch_verify_dns_packet:
 }
 
 /* Make the actual answer for a synthetic "not there" reply */
-unsigned char *make_synth_not_there_answer(unsigned char *a, int count) {
+unsigned char *make_synth_not_there_answer(unsigned char *a, int *count,
+                int type) {
         /* This is the answer for a "not there" reply */
         unsigned char not_there[41] =
         "\xc0\x0c" /* Name */
@@ -773,14 +912,25 @@ unsigned char *make_synth_not_there_answer(unsigned char *a, int count) {
         unsigned char *answer = 0;
         int counter = 0;
 
-        answer = malloc(count + 43);
+        answer = malloc(*count + 43);
+        if(answer == 0) {
+                return 0;
+        }
+
+        if(type == 1) { /* Special case: Return just synth "not there" */
+                for(counter = 0; counter < 40; counter++) {
+                        answer[counter] = not_there[counter];
+                }
+                return answer;
+        }
+
         /* Copy the header they sent us to our reply */
-        for(counter = 0; counter < 12 && counter < count; counter++) {
+        for(counter = 0; counter < 12 && counter < *count; counter++) {
                 answer[counter] = a[counter];
         }
 
         /* Copy the question over to the reply */
-        for(;counter < 520 && counter < count; counter++) {
+        for(;counter < 520 && counter < *count; counter++) {
                 if(a[counter] == 0) {
                         break; /* Quick and dirty "end of name"; yes, I
                                 * check in dw_get_dname_type() to make sure
@@ -788,20 +938,20 @@ unsigned char *make_synth_not_there_answer(unsigned char *a, int count) {
                 }
                 answer[counter] = a[counter];
         }
-        if(count < counter + 5 || counter > 512) { /* Sanity check */
+        if(*count < counter + 5 || counter > 512) { /* Sanity check */
                 free(answer);
                 return 0;
         }
 
         /* Add the rest of the question */
-        count = counter + 5;
-        for(;counter < count; counter++) {
+        *count = counter + 5;
+        for(;counter < *count; counter++) {
                 answer[counter] = a[counter];
         }
 
         /* Add the SOA reply to the answer */
         for(counter = 0; counter < 40; counter++) {
-                answer[count + counter] = not_there[counter];
+                answer[*count + counter] = not_there[counter];
         }
 
         /* Return the answer */
@@ -818,7 +968,7 @@ void make_synth_not_there(int b, SOCKET sock, unsigned char *a, int count) {
         }
 
         /* Copy original header and question in to answer */
-        answer = make_synth_not_there_answer(a,count);
+        answer = make_synth_not_there_answer(a,&count,0);
         if(answer == 0) {
                 return;
         }
@@ -879,13 +1029,16 @@ int send_reply_from_cache(unsigned char *a, ssize_t count, int b, int l) {
         if (rem[b].local[l]->glueless_type != 0) {
                 conn_num = rem[b].local[l]->glueless_conn;
                 if(rem[b].local[l]->glueless_type == 1) { /* Glueless NS */
-                        if(rem[conn_num].child_id == b) {
+                        if(rem[conn_num].child_id == b) { /* Sanity check */
                                 dwx_glueless_done(query, conn_num);
                         }
                         ret = 4; /* Yes, we want to end the connection */
                 } else if(rem[b].local[l]->glueless_type == 2) { /* CNAME */
                         dwx_incomplete_cname_done(query, b, l);
                         ret = 4; /* End connection */
+                } else if(rem[b].local[l]->glueless_type == 3) {/* Saw CNAME */
+                        dwx_cached_cname_done(query, b, l, 0);
+                        ret = 4; /* End connection */
                 }
                 goto catch_send_reply_from_cache;
         } else if (rem[b].local[l]->ip.len == 4) {
@@ -905,7 +1058,8 @@ int send_reply_from_cache(unsigned char *a, ssize_t count, int b, int l) {
         }
 
         get_reply_from_cache(query,&client,rem[b].local[l]->from_socket,
-                        rem[b].local[l]->local_id, 0,-1);
+                        rem[b].local[l]->local_id, 0,-1,
+                        rem[b].local[l]->orig_query, 0);
 
 catch_send_reply_from_cache:
         if(query != 0) {
@@ -920,6 +1074,7 @@ int get_rem_udp_packet_core(unsigned char *a, ssize_t count,
                 int b, SOCKET sock, int l) {
 
         int cache_dns_reply_return_value = -1;
+        int z = 0, y = 0, x = 0;
 
         if(count < 12) {
                 return -1; /* Not a DNS packet at all */
@@ -929,9 +1084,69 @@ int get_rem_udp_packet_core(unsigned char *a, ssize_t count,
                 return -1;
         }
 
+        /* answers.yahoo.com hack: A truncated reply upstream with
+         * answers will be treated as if it has only one answer */
+        if((a[2] & 0x02) != 0x00 && a[7] > 0 &&
+                        key_n[DWM_N_truncation_hack] == 1) {
+                a[6] = a[8] = a[9] = a[10] = a[11] = 0;
+                a[7] = 1;
+                /* Cut off answer after first reply */
+                for(z=12;z<count;z++) {
+                        if(y == 0) { /* Question Length label */
+                                y = 1;
+                                x = a[z];
+                                if(x > 63) {
+                                        return -1; /* funny packet */
+                                }
+                                if(x == 0) {
+                                        x = 4;
+                                        y = 7;
+                                }
+                        } else if(y == 1) { /* In question label */
+                                x--;
+                                if(x == 0) {
+                                        y = 0;
+                                }
+                        } else if(y == 7) { /* Type/class of question */
+                                x--;
+                                if(x == 0) {
+                                        y = 2;
+                                }
+                        } else if(y == 2) { /* First part of name */
+                                if(a[z] != 0xc0) {
+                                        return -1; /* funny packet */
+                                }
+                                y = 3;
+                        } else if(y == 3) { /* Second part of name */
+                                if(a[z] != 0x0c) {
+                                        return -1; /* funny packet */
+                                }
+                                y = 4;
+                                x = 8;
+                        } else if(y == 4) { /* Answer Type, class, or TTL */
+                                x--;
+                                if(x == 0) {
+                                        y = 5;
+                                }
+                        } else if(y == 5) {
+                                if(a[z] != 0) {
+                                        return -1; /* RDLENGTH too long */
+                                }
+                                y = 6;
+                        } else if(y == 6) {
+                                if(z + a[z] + 1> count) {
+                                        return -1; /* shouldn't happen */
+                                }
+                                count = z + a[z] + 1;
+                                y = 8; /* We're done */
+                        }
+                }
+                a[2] &= 0xfd;
+        }
+
         if((a[2] & 0x02) == 0x00) { /* If not truncated */
                 fflush(stdout);
-                cache_dns_reply_return_value = cache_dns_reply(a,count,b);
+                cache_dns_reply_return_value = cache_dns_reply(a,count,b,0);
                 if(cache_dns_reply_return_value == -2) { /* Make synth NX */
                         make_synth_not_there(b,sock,a,count);
                         return -1; /* Bad reply and they got a Synth NX */
@@ -950,8 +1165,11 @@ int get_rem_udp_packet_core(unsigned char *a, ssize_t count,
                         rem[b].local[l]->local_id,b,rem[b].is_upstream);
                 /* Give the UDP connection more time before timeout, so
                  * we can fully process the TCP connection */
-                rem[b].die = get_time() + ((int64_t)timeout_seconds << 10);
+                rem[b].die = get_time() + ((int64_t)timeout_seconds << 11);
                 return 2; /* Don't kill pending UDP connection */
+        } else { /* Truncated over UDP; just given them a blank "truncated"
+                  * reply */
+                cache_dns_reply(a,count,b,1);
         }
 
         /* Now make sure the ID is the same as the one the client
